(* Semantic checking for the MicroC compiler *)

open Ast
open Sast

module StringMap = Map.Make(String)

(* Semantic checking of the AST. Returns an SAST if successful,
   throws an exception if something is wrong.

   Check each global variable, then check each function *)

let check (globals, structs, functions) =

  (* Verify a list of bindings has no void types or duplicate names *)
  let check_binds (kind : string) (binds : bind list) =
    let rec dups = function
        [] -> ()
      |	((_,n1) :: (_,n2) :: _) when n1 = n2 ->
	  raise (Failure ("duplicate " ^ kind ^ " " ^ n1))
      | _ :: t -> dups t
    in dups (List.sort (fun (_,a) (_,b) -> compare a b) binds)
  in

  (**** Check global variables ****)

  check_binds "global" globals;

  (**** Check struct declarations ****)
  (* structs are only ever declared in the global scope *)
  let check_struct (structs : struct_decl list) =

    let rec dups = function
        [] -> ()
      |	((n1,_) :: (n2,_) :: _) when n1 = n2 ->
	  raise (Failure ("duplicate struct " ^ n1))
      | _ :: t -> dups t
    in
    (* Verify no duplicate struct names *)
    dups (List.sort (fun (a,_) (b,_) -> compare a b) structs);
    (* Verify no duplicate struct field names *)
    List.iter (fun (_, field_list) -> dups (List.sort (fun (a,_) (b,_) -> compare a b) field_list)) structs;
  in

  check_struct structs;

  (**** Check functions ****)

  (* Create struct symbol table *)
  let struct_decls =
    let add_struct map (name, fields) = StringMap.add name fields map
    in List.fold_left add_struct StringMap.empty structs
  in

  let find_struct s =
    try StringMap.find s struct_decls
    with Not_found -> raise (Failure ("unrecognized struct " ^ s))
  in

  (* Construct string by joining "field:type_of_expr" of fields in sorted order with ',' *)
  let string_of_fields (field_list : struct_field list) =
    String.concat "," (List.map (fun (field, typ) -> field ^ (string_of_typ typ)) (List.sort (fun (a,_) (b,_) -> compare a b) field_list))
  in
  (* Maps string_of_fields -> struct_name *)
  let struct_decls_rev =
    let add_struct map (name, fields) = StringMap.add (string_of_fields fields) name map
    in List.fold_left add_struct StringMap.empty structs
  in

  (* Check if fields_str exists in struct table, if it does return the struct name, else return fields_str *)
  let get_struct_id fields_str =
    try StringMap.find fields_str struct_decls_rev
    with Not_found -> fields_str
  in

  (* Initialize function symbol table with built-in functions: no bodies *)
  let built_in_fn_decls =
    let add_bind map (name, args, ret_type) = StringMap.add name {
      typ = ret_type;
      fname = name;
      formals = args;
      locals = []; body = [] } map
    in List.fold_left add_bind StringMap.empty [
      ("printi", [(Int, "arg")], Int);
      ("printb", [(Bool, "arg")], Int);
      ("prints", [(String, "arg")], Int);
      ("printc", [(Char, "arg")], Int);
      ("printd", [(Double, "arg")], Int);
      ("printbig", [(Int, "arg")], Int);
      (* following args have non-native types, refer to Call rule *)
      ("lena", [(Int, "arg")], Int);                    (* arg type: array *)
      ("lenm", [(Int, "arg")], Int);                    (* arg type: map *)
      ("lens", [(String, "arg")], Int);                 (* arg type: string *)
      ("getKeys", [(Int, "map")], Int);                 (* arg type: map, ret type: array *)
      ("map_init", [], Int);
      ("arr_init", [], Int);
      ("graph_init", [], Int);
      ("getSrc", [(Int, "edge")], Int);                 (* arg type: edge; ret type: struct *)
      ("getDst", [(Int, "edge")], Int);                 (* arg type: edge; ret type: struct *)
      ("getVal", [(Int, "edge")], Int);                 (* arg type: edge; ret type: struct *)
      ("setSrc", [(Int, "edge"); (Int, "src")], Int);   (* arg types: edge, src; ret type: struct *)
      ("setDst", [(Int, "edge"); (Int, "dst")], Int);   (* arg type: edge, dst; ret type: struct *)
      ("setVal", [(Int, "edge"); (Int, "val")], Int);   (* arg type: edge, val; ret type: struct *)
      ("append", [(Int, "arr"); (Int, "elmt")], Int);
      ("addEdge", [(Int, "graph"); (Int, "edge")], Int);(* arg type: graph, edge; ret type: edge *)
      ("getEdges", [(Int, "graph")], Int);              (* arg type: graph, ret type: array of edges *)
      ("getNodes", [(Int, "graph")], Int);              (* arg type: graph, ret type: array of nodes *)
      ("int_of_double", [(Double, "arg")], Int);
      ("double_of_int", [(Int, "arg")], Double);
  ]
  in

  (* Add function names to symbol table *)
  let add_func map fd =
    let built_in_err = "function " ^ fd.fname ^ " may not be defined"
    and dup_err = "duplicate function " ^ fd.fname
    and make_err er = raise (Failure er)
    and n = fd.fname (* Name of the function *)
    in match fd with (* No duplicate functions or redefinitions of built-ins *)
         _ when StringMap.mem n built_in_fn_decls -> make_err built_in_err
       | _ when StringMap.mem n map -> make_err dup_err
       | _ ->  StringMap.add n fd map
  in

  (* Collect all function names into one symbol table *)
  let function_decls = List.fold_left add_func built_in_fn_decls functions
  in

  (* Return a function from our symbol table *)
  let find_func s =
    try StringMap.find s function_decls
    with Not_found -> raise (Failure ("unrecognized function " ^ s))
  in

  let _ = find_func "main" in (* Ensure "main" is defined *)

  let check_function func =
    (* Make sure no formals or locals are void or duplicates *)
    check_binds "formal" func.formals;
    check_binds "local" func.locals;

    (* Raise an exception if the given rvalue type cannot be assigned to
       the given lvalue type *)
    let check_assign lvaluet rvaluet err =
       if lvaluet = rvaluet then lvaluet else raise (Failure err)
    in

    (* Build local symbol table of variables for this function *)
    let symbols = List.fold_left (fun m (ty, name) -> StringMap.add name ty m)
	                StringMap.empty (globals @ func.formals @ func.locals )
    in

    (* Return a variable from our local symbol table *)
    let type_of_identifier s =
      try StringMap.find s symbols
      with Not_found -> raise (Failure ("undeclared identifier " ^ s))
    in

    (* Helper functions *)
    let rec check_all_same_types l err_type = match l with
        [] -> ()
        |	(t1 :: t2 :: _) when t1 <> t2 ->
                raise (Failure ("non-duplicate " ^ err_type ^ ": t1 = " ^ (string_of_typ t1) ^ ", t2 = " ^ (string_of_typ t2)))
        | _ :: t -> check_all_same_types t err_type
    in
    let rec check_all_diff_vals l err_type = match l with
        [] -> ()
        |	(v1 :: v2 :: _) when v1 = v2 ->
                raise (Failure ("duplicate " ^ err_type ^ ": v1 = v2 = " ^ (string_of_expr v1)))
        | _ :: t -> check_all_diff_vals t err_type
    in

    (* Return a semantically-checked expression, i.e., with a type *)
    let rec expr = function
        IntLit  l -> (Int, SIntLit l)
      | DubLit l -> (Double, SDubLit l)
      | CharLit l -> (Char, SCharLit l) (*TODO: may have error here?*)
      | StrLit l -> (String, SStrLit l)
      | StructLit l ->
        (* Verify no duplicate struct field names *)
        check_all_diff_vals (List.sort (fun (a) (b) -> compare a b) (List.map (fun (field, _) -> StrLit(field)) structs)) "struct fields";
        (* Get a struct_field list by using expr *)
        let get_struct_field_list (struct_lit : (string * expr) list) =
          List.map (fun (a,b) -> (a, fst (expr b))) struct_lit
        in
        (* construct string_of_fields *)
        let get_struct_type struct_lit =
          get_struct_id (string_of_fields (get_struct_field_list struct_lit))
        in
        (* construct semantically checked lit *)
        let get_sstruct_lit struct_lit =
          List.map (fun (a,b) -> (a, expr b)) struct_lit
        in
        let sstruct_lit = get_sstruct_lit l
        in
        (* Verify that all expr are non-container types *)
        let check_nc_types = function
            Int -> ()
          | Bool -> ()
          | Double -> ()
          | Char -> ()
          | String -> ()
          | StructType(_) -> ()
          | _ as t -> raise (Failure ("container type found inside struct literal " ^ string_of_typ t))
        in
        List.iter check_nc_types (List.map snd (get_struct_field_list l));
        (* Return (StructType(StructId), SStructLit l)*)
        (StructType(get_struct_type l), SStructLit sstruct_lit)
      | MapLit l ->
        let get_key_types (map_lit : (expr * expr) list) =
          List.map (fun (k, _) -> fst (expr k)) map_lit
        in
        let get_key_values (map_lit : (expr * expr) list) =
          List.map (fun (k, _) -> k) map_lit
        in
        let get_value_types (map_lit : (expr * expr) list) =
          List.map (fun (_, v) -> fst (expr v)) map_lit
        in
        (* Verify all keys are same type *)
        check_all_same_types (get_key_types l) "key type";
        (* Verify all values are same type *)
        check_all_same_types (get_value_types l) "value type";
        (* Verify no duplicate keys *)
        check_all_diff_vals (get_key_values l) "keys";
        (* Return (Map(t1, t2), SMapLit(sl)) *)
        let t1 = List.nth (get_key_types l) 0 in
        let t2 = List.nth (get_value_types l) 0 in
        let sl = List.map (fun (k, v) -> (expr k, expr v)) l in
        (Map(t1, t2), SMapLit sl)
      | GraphLit l ->
        let get_node_types (graph_lit : expr list) =
          List.map (fun (e) -> match fst (expr e) with
              Edge (nt, _) -> nt
            | _ as t -> raise (Failure ("Expected Edge, received " ^ string_of_typ t))
          ) graph_lit
        in
        let get_edge_types (graph_lit : expr list) =
          List.map (fun (e) -> match fst (expr e) with
              Edge (_, et) -> et
            | _ as t -> raise (Failure ("Expected Edge, received " ^ string_of_typ t))
          ) graph_lit
        in
        (* Verify node types are the same for each EdgeLit *)
        check_all_same_types (get_node_types l) "node type";
        (* Verify edge types are the same for each EdgeLit *)
        check_all_same_types (get_edge_types l) "edge type";
        let sl = List.map (fun (a) -> (expr a)) l in
        let nt = match fst (List.nth sl 0) with
            Edge (nt, _) -> nt
          | _ as t -> raise (Failure ("Expected Edge, received " ^ string_of_typ t))
        in
        let et = match fst (List.nth sl 0) with
            Edge (_, et) -> et
          | _ as t -> raise (Failure ("Expected Edge, received " ^ string_of_typ t))
        in
        (* Return (Graph(t1, t2), SGraphLit sl) *)
        (Graph(nt, et), SGraphLit sl)
      | EdgeLit (node1, node2, edge) ->
        (* Verify node types are the same *)
        let (sn1, sn2, se) = (expr node1, expr node2, expr edge)
        in
        let verify_node_types = function
            (snode1, snode2, _) when (fst snode1) = (fst snode2) -> ()
          | (snode1, snode2, _) when (fst snode1) != (fst snode2) ->
              raise (Failure ("node types not the same: " ^ (string_of_typ (fst snode1)) ^ " and " ^ (string_of_typ (fst snode2))))
          | _ -> raise (Failure ("this should not happen"))
        in
        verify_node_types (sn1, sn2, se);
        (* Return (Edge(t1, t2), SEdgeLit sl) *)
        (Edge(fst sn1, fst se), SEdgeLit (sn1, sn2, se))
      | ArrayLit l ->
          let get_array_types (array_lit : expr list) =
            List.map (fun (a) -> (fst (expr a))) array_lit
          in
          let check_typ = function
              Int -> ()
            | Double -> ()
            | Bool -> ()
            | Char -> ()
            | String -> ()
            | StructType(_) -> ()
            | Edge(_, _) -> ()
            | _ as t -> raise (Failure ("illegal array type: " ^ (string_of_typ t)))
          in
          List.iter check_typ (get_array_types l);
          check_all_same_types (get_array_types l) "array element type";
          let a1 = List.nth (get_array_types l) 0 in
          let sl = List.map (fun (a) -> (expr a)) l in
          (Array(a1), SArrayLit sl)
      | True  -> (Bool, STrue)
      | False -> (Bool, SFalse)
      | Noexpr     -> (Int, SNoexpr)
      | Id s       -> (type_of_identifier s, SId s)
      | Index (s, e) -> (
        let (typ, se) = expr e in
        match (type_of_identifier s) with
          Array(t)    -> (match typ with
            Int -> (t, SArrayIndex(s, (typ, se)))
          | _ -> raise (Failure ("array index must be int, received " ^ (string_of_typ typ)))
          )
        | Map(kt, vt) -> (match typ with
            String | Int | Char | StructType(_) when typ = kt -> (vt, SMapIndex(s, expr e))
          | _ -> raise (Failure ("map index must match key type, received " ^ (string_of_typ typ) ^ " but expected " ^ (string_of_typ kt)))
          )
        | _ as t -> raise (Failure ("illegal index operation: access " ^ (string_of_typ (fst (expr e))) ^ " from " ^ (string_of_typ t)))
      )
      | StructFieldAccess (s, t) -> (
        (* verify struct field is a valid field of the struct *)
        match (type_of_identifier s) with
            StructType(r) -> (
                 let field_list = find_struct r in
                 let found = List.find (fun a -> fst a = t) field_list in
                 ((snd found), SStructFieldAccess(s, t))
            )
          | _ -> raise (Failure ( s ^ " is not a struct type"))
      )
      | Assign(l, r) as ex ->
        let (lt, _) = expr l
        and (rt, _) = expr r in
        let err = "illegal assignment " ^ string_of_typ lt ^ " != " ^
          string_of_typ rt ^ " in " ^ string_of_expr ex in
        let check_for_inits = match lt with
            Map(a,b) ->
              let match_r = match r with
                  Call("map_init", []) ->
                    (Map(a,b), SAssign(expr l, (Map(a,b), (SCall("map_init", [])))))
                | _ -> (check_assign lt rt err, SAssign(expr l, expr r))
              in match_r
          | Graph(a,b) ->
              let match_r = match r with
                  Call("graph_init", []) ->
                    (Graph(a,b), SAssign(expr l, (Graph(a,b), (SCall("graph_init", [])))))
                | _ -> (check_assign lt rt err, SAssign(expr l, expr r))
              in match_r
          | Array(a) ->
              let match_r = match r with
                  Call("arr_init", []) ->
                    (Array(a), SAssign(expr l, (Array(a), (SCall("arr_init", [])))))
                | _ -> (check_assign lt rt err, SAssign(expr l, expr r))
              in match_r
          | _ -> (check_assign lt rt err, SAssign(expr l, expr r))
        in check_for_inits;
      | Reassign(l, r) as ex ->
        let (lt, _) = expr l
        and (rt, _) = expr r in
        let err = "illegal assignment " ^ string_of_typ lt ^ " != " ^
          string_of_typ rt ^ " in " ^ string_of_expr ex
        in (check_assign lt rt err, SAssign(expr l, expr r))
      | Unop(op, e) as ex ->
          let (t, e') = expr e in
          let ty = match op with
            Neg when t = Int || t = Double -> t
          | Not when t = Bool -> Bool
          | Incr when t = Int -> Int
          | Decr when t = Int -> Int
          | _ -> raise (Failure ("illegal unary operator " ^
                                 string_of_uop op ^ string_of_typ t ^
                                 " in " ^ string_of_expr ex))
          in (ty, SUnop(op, (t, e')))
      | Binop(e1, op, e2) as e ->
          let (t1, e1') = expr e1
          and (t2, e2') = expr e2 in
          (* All binary operators require operands of the same type *)
          let same = t1 = t2 in
          (* Determine expression type based on operator and operand types *)
          let ty = match op with
            Add | Sub | Mult | Div | Mod      when same && t1 = Int    -> Int
          | Add                               when same && t1 = String -> String
          | DAdd | DSub | DMult | DDiv | DMod when same && t1 = Double -> Double
          | Equal | Neq                       when same                -> Bool
          | Less | Leq | Greater | Geq
                                when same && (t1 = Int || t1 = Double) -> Bool
          | And | Or                            when same && t1 = Bool -> Bool
          | In when (
            match t2 with
              (* Array: right side is array and left side is array element *)
              Array t1' -> t1' = t1
              (* Map: right side is map and left side is key *)
              | Map (kt, _) -> kt = t1
              (* Graph: right side is graph and left side is struct (node) lit or edge lit *)
              | Graph (_nt, _et) -> (
                  match t1 with
                    StructType _ -> _nt = t1
                  | Edge (nt', et') -> nt' = _nt && et' = _et
                  | _ as t -> raise (Failure ("expected struct or edge, received " ^ string_of_typ t))
                )
              | _ -> false
          )
          -> Bool
          | _ -> raise (
	      Failure ("illegal binary operator " ^
                       string_of_typ t1 ^ " " ^ string_of_op op ^ " " ^
                       string_of_typ t2 ^ " in " ^ string_of_expr e))
          in (ty, SBinop((t1, e1'), op, (t2, e2')))
      | Call(fname, args) as call ->
          let fd = find_func fname in
          let param_length = List.length fd.formals in
          if List.length args != param_length then
            raise (Failure ("expecting " ^ string_of_int param_length ^
                            " arguments in " ^ string_of_expr call))
          else match fname with
            "lena" ->
              let (at, ae) = expr (List.nth args 0) in
              let at = match at with
                  Array(_) -> at
                | _ -> raise (Failure ("expecting array type, received " ^ string_of_typ at))
              in
              (Int, SCall(fname, [(at, ae)]))
          | "lenm" ->
              let (mt, me) = expr (List.nth args 0) in
              let mt = match mt with
                  Map(_, _) -> mt
                | _ -> raise (Failure ("expecting map type, received " ^ string_of_typ mt))
              in
              (Int, SCall(fname, [(mt, me)]))
          | "getKeys" ->
              let (mt, me) = expr (List.nth args 0) in
              let (kt, mt) = match mt with
                  Map(kt, _) -> (kt, mt)
                | _ -> raise (Failure ("expecting map type, received " ^ string_of_typ mt))
              in
              (Array(kt), SCall(fname, [(mt, me)]))
          | "getSrc" | "getDst" ->
              let (et, ee) = expr (List.nth args 0) in
              let (nt, et) = match et with
                  Edge(nt, _) -> (nt, et)
                | _ -> raise (Failure ("expecting edge type, received " ^ string_of_typ et))
              in
              (nt, SCall(fname, [(et, ee)]))
          | "getVal" ->
              let (et, ee) = expr (List.nth args 0) in
              let (vt, et) = match et with
                  Edge(_, vt) -> (vt, et)
                | _ -> raise (Failure ("expecting edge type, received " ^ string_of_typ et))
              in
              (vt, SCall(fname, [(et, ee)]))
          | "setSrc" | "setDst" ->
              let (et, ee) = expr (List.nth args 0) in
              let (nt, ne) = expr (List.nth args 1) in
              let verify_node_types = function
                  (Edge(nt', _), nt) when nt' = nt -> ()
                | (Edge(nt', _), nt) -> raise (Failure ("expecting matching node types, received " ^ string_of_typ nt' ^ " and " ^ string_of_typ nt))
                | ((_ as t1), (_ as t2)) -> raise (Failure ("expecting edge and node types, received " ^ string_of_typ t1 ^ " and " ^ string_of_typ t2))
              in
              verify_node_types (et, nt);
              (nt, SCall(fname, [(et, ee); (nt, ne)]))
          | "setVal" ->
              let (et, ee) = expr (List.nth args 0) in
              let (vt, ve) = expr (List.nth args 1) in
              let verify_val_types = function
                  (Edge(_, vt'), vt) when vt' = vt -> ()
                | (Edge(_, vt'), vt) -> raise (Failure ("expecting matching value types, received " ^ string_of_typ vt' ^ " and " ^ string_of_typ vt))
                | ((_ as t1), (_ as t2)) -> raise (Failure ("expecting edge and edge value types, received " ^ string_of_typ t1 ^ " and " ^ string_of_typ t2))
              in
              verify_val_types (et, vt);
              (vt, SCall(fname, [(et, ee); (vt, ve)]))
          | "append" ->
              let (at, ae) = expr (List.nth args 0) in
              let (vt, ve) = expr (List.nth args 1) in
              let verify_arr_type = match at with
                    Array(_) -> (match vt with
                        | Int -> ()
                        | Double -> ()
                        | Bool -> ()
                        | Char -> ()
                        | String -> ()
                        | StructType _ -> ()
                        | Edge (_, _) -> ()
                        | _ -> raise (Failure("incompatible type"))
                    )
                    | _ -> raise (Failure("can only append to array"))
              in
              verify_arr_type;
              (vt, SCall(fname, [(at, ae); (vt, ve)]))
          | "addEdge" ->
              let (gt, ge) = expr (List.nth args 0) in
              let (et, ee) = expr (List.nth args 1) in
              let verify_edge_types = function
                  (Graph(nt, vt), Edge(nt', vt')) when nt = nt' && vt = vt' -> ()
                | (Graph(nt, vt), Edge(nt', vt')) -> raise (Failure ("expecting matching node and value types, received " ^ string_of_typ nt ^ ", " ^ string_of_typ nt' ^ ", " ^ string_of_typ vt ^ ", " ^ string_of_typ vt'))
                | ((_ as t1), (_ as t2)) -> raise (Failure ("expecting graph and edge types, received " ^ string_of_typ t1 ^ " and " ^ string_of_typ t2))
              in
              verify_edge_types (gt, et);
              (et, SCall(fname, [(gt, ge); (et, ee)]))
          | "getEdges" ->
              let (gt, ge) = expr (List.nth args 0) in
              let et = match gt with
                  Graph(nt, vt) -> Array(Edge(nt, vt))
                | _ -> raise (Failure ("expecting graph type, received " ^ string_of_typ gt))
              in
              (et, SCall(fname, [(gt, ge)]))
          | "getNodes" ->
              let (gt, ge) = expr (List.nth args 0) in
              let et = match gt with
                  Graph(nt, _) -> Array(nt)
                | _ -> raise (Failure ("expecting graph type, received " ^ string_of_typ gt))
              in
              (et, SCall(fname, [(gt, ge)]))
          | _ ->
              let check_call (ft, _) e =
                let (et, e') = expr e in
                let err = "illegal argument found " ^ string_of_typ et ^
                  " expected " ^ string_of_typ ft ^ " in " ^ string_of_expr e
                in (check_assign ft et err, e')
              in
              let args' = List.map2 check_call fd.formals args
              in (fd.typ, SCall(fname, args'))
    in

    let check_bool_expr e =
      let (t', e') = expr e
      and err = "expected Boolean expression in " ^ string_of_expr e
      in if t' != Bool then raise (Failure err) else (t', e')
    in

    (* Return a semantically-checked statement i.e. containing sexprs *)
    let rec check_stmt = function
        Expr e -> SExpr (expr e)
      | If(p, b1, b2) -> SIf(check_bool_expr p, List.map check_stmt b1, List.map check_stmt b2)
      | IfElseifs (p, b1, l, b3) -> SIfElseifs(check_bool_expr p, List.map check_stmt b1, (List.map (fun(_, b2) -> check_bool_expr p, List.map check_stmt b2) l), List.map check_stmt b3)
      | For(e1, e2, e3, st) ->
	  SFor(expr e1, check_bool_expr e2, expr e3, List.map check_stmt st)
      | While(p, s) -> SWhile(check_bool_expr p, List.map check_stmt s)
      | Return e -> let (t, e') = expr e in
        if t = func.typ then SReturn (t, e')
        else raise (
	  Failure ("return gives " ^ string_of_typ t ^ " expected " ^
		   string_of_typ func.typ ^ " in " ^ string_of_expr e))

	    (* A block is correct if each statement is correct and nothing
	       follows any Return statement.  Nested blocks are flattened. *)
      | Block sl ->
          let rec check_stmt_list = function
              [Return _ as s] -> [check_stmt s]
            | Return _ :: _   -> raise (Failure "nothing may follow a return")
            | Block sl :: ss  -> check_stmt_list (sl @ ss) (* Flatten blocks *)
            | s :: ss         -> check_stmt s :: check_stmt_list ss
            | []              -> []
          in SBlock(check_stmt_list sl)
      | _ as s -> raise (Failure ("unexpected statement: " ^ (string_of_stmt s)))

    in (* body of check_function *)
    { styp = func.typ;
      sfname = func.fname;
      sformals = func.formals;
      slocals  = func.locals;
      sbody = match check_stmt (Block func.body) with
	SBlock(sl) -> sl
      | _ -> raise (Failure ("internal error: block didn't become a block?"))
    }
  in (globals, structs, List.map check_function functions)
